namespace FBPLib
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Reflection;
    using System.IO;
    using System.Threading;
    using Spring.Threading.Helpers;   // CountDownLatch
    //  using System.Text.RegularExpressions;
    using FBP.Properties;

    /// <summary> The abstract class which all flow Networks extend directly or
    /// indirectly.  A specific flow Network must override the <code>define()</code>
    /// method, which is written using the <i>mini-language</i>
    /// (actually just highly restricted Java invoking the <code>protected</code>
    /// methods of this class).  The mini-language specifies what
    /// threads are to be created using which components, and what
    /// connections are established between the ports of those components.
    /// Here is a sample: 
    /// <pre>
    /// connect(component("Read", typeof(ReadText)),
    ///		port("OUT"),
    ///		component("Write", typeof(WriteText)),
    ///		port("IN"));
    /// Object d = (Object) "Data\\Testing.txt";
    /// initialize(d,
    ///		component("Read"),
    ///		port("SOURCE"));
    /// Stream st = Console.OpenStandardOutput();
    ///	initialize(st,
    ///		component("Write"),
    ///		port("DESTINATION"));
    /// </pre>
    /// The effect of the above is to create two threads named <code>Read</code>
    /// (an instance of <code>ReadText</code>) and <code>Write</code> (an instance of
    /// <code>WriteText</code>, connect the <code>OUT</code> port of the former to
    /// the <code>IN</code> port of the latter, and initialize the <code>SOURCE</code>
    /// and <code>DESTINATION</code> parameter ports with suitable Java objects
    /// for reading the standard input and writing the standard output.
    /// *
    /// </summary>

    public abstract class Network : Component
    {
        /* *
           * Copyright 2007, 2008, 2009, J. Paul Morrison.  At your option, you may copy, 
           * distribute, or make derivative works under the terms of the Clarified Artistic License, 
           * based on the Everything Development Company's Artistic License.  A document describing 
           * this License may be found at http://www.jpaulmorrison.com/fbp/artistic2.htm. 
           * THERE IS NO WARRANTY; USE THIS PRODUCT AT YOUR OWN RISK.
           * */


        protected internal static int DEBUGSIZE = 1;

        protected internal static int PRODUCTIONSIZE = 10;

        //internal static int _defaultCapacity = DEBUGSIZE;       // change this when you go to production
        internal static int _defaultCapacity = PRODUCTIONSIZE; // use this one for production

        //UPGRADE_NOTE: The initialization of  'components' was moved to method 'InitBlock'. 'ms-help://MS.VSCC/commoner/redir/redirect.htm?keyword="jlca1005"'
        internal Dictionary<string, Component> _components;

        internal bool _active = true;   // for deadlock detection   

        internal Dictionary<Component, TimeoutHandler> _timeouts = new Dictionary<Component, TimeoutHandler>();

        Thread _mainthread = null;

        public CountDownLatch _cdl;

        //public static Network _topNetwork;

        //static internal string _netname;

        public object[] parms;

        string _strPath;

        public bool _deadlock;

        private Exception _error;
        bool _abort = false;

        public bool _deadlockTest = true; // false if SubNet
        List<String> msgs = null;
        public static bool _tracing;
        public static String _tracePath;
        private StreamWriter _traceWriter;
        public bool _useConsole = false;
        public static bool _forceConsole = false;
        public static List<StreamWriter> _traceFileList = null;

        public new void InitBlock()
        {
            _components = new Dictionary<string, Component>();
            //_network = this;

            // Get the directory our exe is running from.
            _strPath = System.Reflection.Assembly.GetExecutingAssembly().Location;
            _strPath = _strPath.Substring(0, _strPath.LastIndexOf('\\'));
            Settings s = Settings.Default;
            _tracing = s.Tracing;  // get value from Properties for FBPLib
            _tracePath = @"C:\Temp\";

            if (!(this is SubNet))  // i.e. Network only!
            {
                int i = Name.LastIndexOf(".");
                if (i != -1)
                    Name = Name.Substring(i + 1);
                _traceFileList = new List<StreamWriter>();
            }
        }
        // return object to use for synchronising state changes
        internal object StateSynchRoot
        {
            get { return (_components as ICollection).SyncRoot; }
        }

        /// <summary>Drive define method of Network
        /// </summary>

        internal virtual void CallDefine()
        {
            // don't turn every exception into a FlowError
            //try
            //{
            Define();
            //}
            //catch (System.Exception t)
            //{
            //     FlowError.Complain(t.Message);
            // }
        }

        /// <summary>Returns a Component class object, given the Component name in the Network
        /// </summary>		
        protected internal Component Component(string name)
        {
            if (!_components.ContainsKey(name))
                FlowError.Complain("Reference to unknown Component " + name);
            Component c = (Component)(_components[name]);

            c.Name = name;
            return c;
        }
        /// <summary>Stores the Component class object with its Network name in the Hashtable called
        /// 'components'
        /// </summary>		
        protected internal Component Component(string name, System.Type type)
        {
            if (_components.ContainsKey(name))
                FlowError.Complain("Attempt to redefine Component " + name);
            Component c = null;
            //UPGRADE_NOTE: Exception 'java.lang.InstantiationException' was converted to ' ' which has different behavior. 'ms-help://MS.VSCC/commoner/redir/redirect.htm?keyword="jlca1100"'
            Type foundType = null;
            // Search for all DLLs in the folder 
            foreach (string strFile in Directory.GetFiles(_strPath, "*.dll"))
            {
                // Open the class library.
                Assembly engine = Assembly.LoadFrom(strFile);

                Type[] types = engine.GetTypes();
                foreach (Type t in types)
                {
                    if ((t.FullName == type.FullName) & (t.IsClass))
                    {
                        foundType = t;
                        break;
                    }
                }
                if (foundType != null)
                    break;
            }
            if (foundType == null)
                FlowError.Complain("Component not found: " + type.FullName);

            try
            {
                //ConstructorInfo cnst = foundType.GetConstructor(Type.EmptyTypes);
                //c = (Component)(cnst.Invoke(null));
                c = NewComponent(foundType, name, this);
            }
            catch (System.UnauthorizedAccessException e)
            {
                FlowError.Complain("Illegal access to " + type.FullName);
                return null; // unreachable
            }
            catch (System.Exception e)
            {
                FlowError.Complain("Cannot instantiate " + type.FullName);
                return null; // unreachable
            }

            c.Type = foundType;
            c.Name = name;
            Network._network = this;
            c._mother = this;
            c._thread = new Thread(new ThreadStart(c.ThreadMain));
            _components.Add(name, c);
            return c;


        }


        /// <summary>Connect an output port of one Component to an input port
        /// of another, using default capacity
        /// </summary>
        protected internal void Connect(Component sender, Port outPort, Component receiver, Port inPort)
        {
            Connect(sender, outPort, receiver, inPort, 0);
        }
        /// <summary>Connect an output port of one Component to an input port
        /// of another, specifying a connection capacity
        /// </summary>

        protected internal void Connect(Component sender, Port outPort, Component receiver, Port inPort, int size)
        {
            int cap = size;
            if (size == 0)
                cap = _defaultCapacity;

            //string outName = outPort.displayName;
            //string inName = inPort.displayName;
            if (outPort.displayName.Equals("*"))
            {
                outPort.name = "*OUT";
                outPort.displayName = "*OUT";
            }
            if (inPort.displayName.Equals("*"))
            {
                inPort.name = "*IN";
                inPort.displayName = "*IN";
            }

            if (sender._outputPorts.ContainsKey(outPort.displayName))
                FlowError.Complain("Multiple connections to output port:" + sender.Name + "." + outPort.displayName);

            //Regex re = new Regex(@"^(\w+)\[");  // test for square bracket
            //Match m;
            Connection cx = null;

            if (receiver._inputPorts.ContainsKey(inPort.displayName))
                if (!(receiver._inputPorts[inPort.displayName] is Connection))
                    FlowError.Complain("Mixed connection to input port: " +
                        receiver.Name + "." + inPort.displayName);
                else
                {
                    cx = receiver._inputPorts[inPort.displayName] as Connection;
                    goto ok;
                }

            if (inPort.name.Substring(0, 1).Equals("*"))
            {
                cx = new Connection(cap);
                //cx._network = this;
            }
            else
            {
                //m = re.Match(inName);  // does it have square bracket in the name?
                if (inPort.index == -1)    // not array port
                // so name and displayName are same
                {
                    if (!receiver._inputPorts.ContainsKey(inPort.name))
                    {
                        if (!receiver._inputPortAttrs.ContainsKey(inPort.name) ||
                         receiver._inputPortAttrs[inPort.name].arrayPort)
                            FlowError.Complain("Input port not specified correctly in attributes: " +
                                receiver.Name + "." + inPort.name);
                        receiver._inputPortAttrs.Remove(inPort.name);

                        cx = new Connection(cap);
                        //cx._network = this;
                    }
                }
                else                // it is array port
                {
                    string s = inPort.name;
                    if (!receiver._inputPortAttrs.ContainsKey(s) ||
                        !receiver._inputPortAttrs[s].arrayPort)
                        FlowError.Complain("Input port not specified correctly in attributes: " +
                               receiver.Name + "." + inPort.displayName);
                    cx = new Connection(cap);
                    //cx._network = this;
                    receiver._inputPortAttrs[s].used = true;

                }
            }
            cx.SetReceiver(receiver);
            cx.Name = receiver.Name + "." + inPort.displayName;
            receiver._inputPorts.Add(inPort.displayName, cx);

        ok:
            if (size != 0 && size != cap)
                FlowError.Complain("Connection capacity does not agree with previous specification for\n" +
                    receiver.Name + "." + inPort.displayName);

            // Now start handling output port
            OutputPort op = null;

            if (sender._outputPorts.ContainsKey(outPort.displayName))
            {
                op = sender._outputPorts[outPort.displayName];
                cx.BumpSenderCount();
                goto ok2;
            }

            if (!outPort.name.Substring(0, 1).Equals("*"))
            {
                // m = re.Match(outName);  // does it have square bracket in the name?
                if (outPort.index == -1)  // not array port
                {
                    if (!sender._outputPortAttrs.ContainsKey(outPort.name) ||
                         sender._outputPortAttrs[outPort.name].arrayPort)
                        FlowError.Complain("Output port not specified correctly in attributes: " +
                            sender.Name + "." + outPort.displayName);
                    sender._outputPortAttrs.Remove(outPort.name);
                }
                else                   //yes it is array port
                {
                    string s = outPort.name;
                    if (!sender._outputPortAttrs.ContainsKey(s) ||
                        !sender._outputPortAttrs[s].arrayPort)
                        FlowError.Complain("Output port not specified correctly in attributes: " +
                               sender.Name + "." + outPort.displayName);
                    sender._outputPortAttrs[s].used = true;
                }
            }

            op = new OutputPort();
            sender._outputPorts.Add(outPort.displayName, op);
            cx.BumpSenderCount();
            op._connected = true; // new addition - see discussion w. D. Child

        ok2:
            op._cnxt = cx;
            op.SetSender(sender);
            op.Name = sender.Name + "." + outPort.displayName;
        }

        protected internal void Connect(Component sender, Port outPort, string receiver)
        {
            string[] parts = CPSplit(receiver);
            Connect(sender, outPort, Component(parts[0]), Port(parts[1]), 0);
        }
        protected internal void Connect(Component sender, Port outPort, string receiver, int size)
        {
            string[] parts = CPSplit(receiver);
            Connect(sender, outPort, Component(parts[0]), Port(parts[1]), size);
        }

        protected internal void Connect(string sender, Component receiver, Port inPort)
        {
            string[] parts = CPSplit(sender);
            Connect(Component(parts[0]), Port(parts[1]), receiver, inPort, 0);
        }
        protected internal void Connect(string sender, Component receiver, Port inPort, int size)
        {
            string[] parts = CPSplit(sender);
            Connect(Component(parts[0]), Port(parts[1]), receiver, inPort, size);
        }

        protected internal void Connect(string sender, string receiver)
        {
            Connect(sender, receiver, 0);
        }

        protected internal void Connect(string sender, string receiver, int size)
        {
            string[] sParts, rParts;
            sParts = CPSplit(sender);
            rParts = CPSplit(receiver);

            Connect(Component(sParts[0]), Port(sParts[1]),
               Component(rParts[0]), Port(rParts[1]), size);
        }

        // splits a string into component part and port part

        string[] CPSplit(string s)
        {
            int i = s.IndexOf(".");
            if (i < 0)
            {
                FlowError.Complain("Invalid receiver string: " + s);
            }
            string[] p = { s.Substring(0, i), s.Substring(i + 1) };
            return p;
        }

        public abstract void Define();

        /// <summary>Execute method used by Network being used as if it were a Component
        /// </summary>
        public override void Execute()
        {
            // overridden by specific networks
        }


        /// <summary>Execute Network as a whole
        /// </summary>
        [MTAThreadAttribute]   // this the default, so not really necessary
        public void Go()
        {
            Type t = this.GetType();
            Name = t.FullName;

            _network = this;

            DateTime now = DateTime.Now;

            InitBlock();
            _mainthread = new Thread(delegate()
            {
                try
                {
                    CallDefine();

                    _cdl = new CountDownLatch(_components.Count);


                    Trace(Name + ": run started");
                    _active = true;

                    Initiate();

                    WaitForAll();
                }
                catch (FlowError e)
                {
                    string s = "Flow Error :" + e;

                    Console.Out.WriteLine("Network: " + s);
                    Console.Out.Flush();
                    // rethrow the exception for external error handling
                    // in case of a deadlock: deadlock is the cause
                    throw e;
                }

                if (_error != null)
                {
                    // throw the exception which caused the network to stop
                    throw _error;
                }

                TimeSpan duration = DateTime.Now - now;

                Console.Out.WriteLine("{0} - run time: {1}", Name, duration);
                CloseTraceFiles();

            });

            _mainthread.Start();
        }

        protected internal void Initialize(object param, string receiver)
        {
            string[] parts;
            parts = CPSplit(receiver);
            Initialize(param, Component(parts[0]), Port(parts[1]));
        }
        /// <summary>Build InitializationConnection object
        /// </summary>

        protected internal void Initialize(System.Object content, Component receiver, Port inPort)
        {
            //if (inName.Equals("*"))
            //    inName = "*IN";

            //  Regex re = new Regex(@"^(\w+)\[");
            //  Match m;
            //Connection cx = null;
            // string inName = inPort.displayName;
            if (receiver._inputPorts.ContainsKey(inPort.displayName))
                FlowError.Complain("IIP port cannot be shared: " +
                    receiver.Name + "." + inPort.displayName);

            //m = re.Match(inName);  // does it have square bracket in the name?
            if (inPort.index == -1)    // not array port
            {
                if (!receiver._inputPorts.ContainsKey(inPort.name))
                {
                    if (!receiver._inputPortAttrs.ContainsKey(inPort.name) ||
                     receiver._inputPortAttrs[inPort.name].arrayPort)
                        FlowError.Complain("Input port not specified correctly in attributes: " +
                            receiver.Name + "." + inPort.name);
                    receiver._inputPortAttrs.Remove(inPort.name);
                }
            }
            else                // array port
            {
                string s = inPort.name;
                if (!receiver._inputPortAttrs.ContainsKey(s) ||
                    !receiver._inputPortAttrs[s].arrayPort)
                    FlowError.Complain("Input port not specified correctly in attributes: " +
                           receiver.Name + "." + inPort.displayName);
                receiver._inputPortAttrs[s].used = true;

            }

            InitializationConnection ic = new InitializationConnection(content, receiver);
            receiver._inputPorts[inPort.displayName] = ic;
            ic.Name = receiver.Name + "." + inPort.displayName;
            //ic.network = this;

        }
        /// <summary>Go through components activating any that can be activated
        /// </summary>
        internal virtual void Initiate()
        {

            foreach (Component c in _components.Values)
            {
                foreach (string s in c._outputPortAttrs.Keys)
                {
                    if (!c._outputPortAttrs[s].used && !c._outputPortAttrs[s].optional)
                        FlowError.Complain("Output port not defined in connect statement: " + c.Name + "." + s);
                }
                c.OpenPorts();
            }
            ArrayList selfStarters = new ArrayList();
            foreach (Component c in _components.Values)
            {
                c.SelfStarting = true;

                foreach (IInputPort port in c._inputPorts.Values)
                {
                    if (port is Connection)
                    {
                        //cnxt.CheckTypes();                        
                        c.SelfStarting = false;
                        break;
                    }
                }

                if (c.SelfStarting)
                {
                    selfStarters.Add(c);
                }
            }

            foreach (Component c in selfStarters)
            {
                c.Activate();
            }
        }

        //  public override Object[] Introspect()
        //  {
        //     return new Object[] {
        // "manages a network"};
        //  }


        /// <summary>method to open ports for subnet
        /// </summary>
        public override void OpenPorts()
        {
            /*
            _inport = OpenInput("IN");
            _outport = OpenOutput("OUT");
            substreamSensitive = OpenInput("SUBSTREAM-SENSITIVE");
             * */
        }


        /// <summary>method to register a port name  
        /// </summary>

        public Port Port(string name)
        {
            if (!name.Equals("*") && !name.Equals("*SUBEND") && -1 != name.IndexOf('*'))
            {
                FlowError.Complain("Stray * in port name " + name);
            }
            Port p = new Port(name, -1);
            //p.displayName = String.Format("{0}", name);
            return p;
        }
        /// <summary>method to register a port (with index) 
        /// </summary>

        protected internal Port Port(string name, int index)
        {
            if (!name.Equals("*") && -1 != name.IndexOf('*'))
            {
                FlowError.Complain("Stray * in port name " + name);
            }
            Port p = new Port(name, index);
            //p.displayName = String.Format("{0}[{1}]", name, index);
            return p;
        }


        /// <summary>Test if Network as a whole has terminated
        /// </summary>

        internal virtual void WaitForAll()
        {
            bool possibleDeadlock = false;
            bool deadlock = false;

            int freq = 500;   // check every .5 second
            Settings s = Settings.Default;
            while (true)
            {
                if (_cdl.Await(new TimeSpan(0, 0, 0, 0, freq)))   // 500 msecs 
                    break;  // if CountDownLatch finished, exit
                // if an error occurred, skip deadlock testing
                if (_error != null)
                {
                    break;
                }

                // if the network was aborted, skip deadlock testing 
                if (_abort)
                {
                    break;
                }
                if (!_deadlockTest)
                {
                    continue;
                }
                if (s.DeadlockTestEnabled)
                {
                    TestTimeouts(freq);
                    if (!_active)  // else interval elapsed
                        if (!possibleDeadlock)
                            possibleDeadlock = true;
                        else
                        {
                            deadlock = true;
                            // well, maybe
                            // so test state of components
                            msgs = new List<String>();
                            msgs.Add("Network has deadlocked");
                            if (ListCompStatus(msgs))
                            {
                                //          interruptAll();
                                foreach (string m in msgs)
                                    Console.Out.WriteLine(m);
                                Console.Out.WriteLine("*** Deadlock detected in Network ");
                                Console.Out.Flush();

                                // terminate the net instead of crashing the application
                                Terminate();
                                // tell the caller a deadlock occurred
                                FlowError.Complain("Deadlock detected in Network");
                                break;
                            }
                            // one or more components haven't started or
                            // are in a long wait
                            deadlock = false;
                            possibleDeadlock = false;
                        }
                    _active = false;
                }
            }

            if (deadlock) InterruptAll();
            else
            {

                foreach (Component c in _components.Values)
                { c._thread.Join(); }

                //if (outputCopier != null)
                //   outputCopier.Join();
            }

        }
        /* not needed any more
        bool DeadlockTest()
        {
            List<string> msgs = new List<string>();
            // Messages are added to list, rather than written directly,
            //   in case it is not a deadlock
            msgs.Add("Network has deadlocked");
            lock (StateSynchRoot)
            {
                foreach (Component comp in _components.Values)
                {
                    if (comp.Status == States.Active ||
                        comp.Status == States.LongWait) return false;
                    msgs.Add(String.Format("--- {1,-13} -- {0}", comp.Name, comp.Status));
                }
                foreach (string m in msgs)
                    Console.Out.WriteLine(m);
                //FlowError.Complain("Deadlock detected");
                System.Diagnostics.Trace.Fail("*** Deadlock detected in Network ");
                return true;
            }
        }
        */

        /**
   * Queries the status of the subnet's components.
   * if deadlock, return true, else return false   
   * 
   * @param msgs the message vector for status lines
   */

        bool ListCompStatus(List<String> msgs)
        {
            lock (this)
            {
                foreach (Component comp in _components.Values)
                {
                    if (comp is SubNet)
                    {
                        SubNet subnet = comp as SubNet;
                        if (!subnet.ListCompStatus(msgs))
                            return false;
                    }
                    else
                    {
                        if (comp._status == States.Active || comp._status == States.LongWait)
                        {
                            return false;
                        }

                        string st = Enum.GetName(typeof(States), comp._status);
                        st = (st + "            ").Substring(0, 13);
                        msgs.Add(String.Format("--- {1}     {0}", Name + "." + comp.Name, st));
                    }
                }


                return true;
            }
        }
        // called by WaitForAll method
        void TestTimeouts(int freq)
        {
            lock (_timeouts)
            {
                foreach (TimeoutHandler t in _timeouts.Values)
                {
                    t.Decrement(freq);
                }
            }
        }
        internal virtual void InterruptAll()
        {

            IDictionaryEnumerator all = _components.GetEnumerator();
            while (all.MoveNext())
            {
                Component c = (Component)(all.Value);
                if (!c.IsTerminated())
                {
                    c._thread.Interrupt();
                    Console.Out.WriteLine(c.Name + " interrupted");
                }

            }
        }
        internal void NotifyTerminated(Component comp)
        {
            // lock (StateSynchRoot)
            // {
            //   foreach (Component comp in _components.Values)
            //   {
            //      if (comp.Status != States.Terminated)
            //          return;
            //  }
            //_mainthread.Interrupt(); 
            lock (comp)
            {
                comp._status = States.Terminated;
            }
            Trace("Terminated: " + comp.Name);
            _cdl.CountDown();
            // }
        }
        internal override void Terminate(States newStatus)
        {
            // prevent deadlock testing, components will be shut down anyway
            _abort = true;
            foreach (Component comp in _components.Values)
            {
                comp.Terminate(newStatus);
            }
        }
        /**
           * Handles errors in the network.
           * @param e the exception which specifies the error
           */
        public virtual void SignalError(Exception e)
        {
            // only react to the first error, the others presumably are inherited errors
            if (_error == null)
            {
                // set the error field to let go() throw the exception
                _error = e;
                // terminate the network's components
                foreach (Component comp in _components.Values)
                {
                    comp.Terminate(States.Error);
                }
            }
        }
        /**
   * Shuts down the network.
   */
        public void Terminate()
        {
            Terminate(States.Terminated);
        }

        /* Generate Trace line */
        internal void Trace(string msg)
        {
            if (_tracing)
            {
                lock (this)
                {
                    DateTime now = DateTime.UtcNow;

                    string dt = now.ToString("yyyy-MM-dd HH:mm:ss:fff");

                    string n = GetTracingName();

                    // forceConsole is used for debugging purposes to force writing to the console
                    // useConsole will be set to true if the trace file could not be opened
                    if (_forceConsole || _useConsole)
                    {
                        lock (_network)
                        {
                            Console.Out.WriteLine(dt + " " + n + ": " + msg);
                            Console.Out.Flush();
                        }
                        return;
                    }
                    FileStream fs = null;
                    if (_traceWriter == null)
                    {
                        try
                        {
                            string s = _tracePath + n + "-fulltrace.txt";
                            fs = new FileStream(s, FileMode.OpenOrCreate, FileAccess.Write);
                            _traceWriter = new StreamWriter(fs);
                        }
                        catch (IOException e)
                        {
                            // file cannot be created or opened - disable tracing
                            // _tracing = false;
                            lock (_network)
                            {
                                Console.Out.WriteLine("Trace file " + fs.Name + " could not be opened - \n" +
                                "   writing to console...");
                                Console.Out.WriteLine(dt + " " + n + ": " + msg);
                                Console.Out.Flush();
                            }
                            _useConsole = true;
                            return;
                        }
                        _traceFileList.Add(_traceWriter);
                    } try
                    {
                        _traceWriter.WriteLine(dt + " " + msg);
                        _traceWriter.Flush();
                        // Thread.Sleep(500);  // for testing
                    }
                    catch (IOException e)
                    {
                        Console.Out.WriteLine("Trace file " + fs.Name + " could not be written - \n" +
                                "   writing to console...");
                        Console.Out.WriteLine(dt + " " + n + ": " + msg);
                        Console.Out.Flush();
                    }

                }
            }
        }

        void CloseTraceFiles()
        {
            foreach (StreamWriter x in _traceFileList)
                x.Close();
        }
        string GetTracingName()
        {
            if (_mother != null)
            {
                // we're in a component or in a subnet
                // recursively build the tracing name
                return _mother.GetTracingName() + "." + Name;
            }
            // we're in the Network class
            return Name; // or "Network"

        }

        internal void Trace(string format, params object[] args)
        {
            Trace(String.Format(format, args));
        }

        /**
  * Sets a new path for trace files. By default the current directory will be
  * used.
  * @param path the trace path to set
  */
        public static void SetTracePath(string path)
        {
            if (path.EndsWith(@"\") || path.Equals(""))
            {
                Network._tracePath = path;
            }
            else
            {
                // append the file name separator if it is missing and the path is not
                // empty
                Network._tracePath = path + @"\";
            }
        }
    }
}
